package dev.fitko.fitconnect.client;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.config.defaults.Environments;
import dev.fitko.fitconnect.api.domain.model.callback.Callback;
import dev.fitko.fitconnect.api.domain.model.event.EventPayload;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.attachment.AttachmentEncryptionIssue;
import dev.fitko.fitconnect.api.domain.model.event.problems.attachment.IncorrectAttachmentAuthenticationTag;
import dev.fitko.fitconnect.api.domain.model.event.problems.attachment.MissingAttachment;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataEncryptionIssue;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataHashMismatch;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataSchemaViolation;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.IncorrectMetadataAuthenticationTag;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.InvalidEventLog;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.MissingAuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.NotExactlyOneSubmitEvent;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionForPickup;
import dev.fitko.fitconnect.api.domain.model.submission.SubmissionsForPickup;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectSubscriberException;
import dev.fitko.fitconnect.api.exceptions.internal.AuthenticationTagsEmptyException;
import dev.fitko.fitconnect.api.exceptions.internal.DecryptionException;
import dev.fitko.fitconnect.api.exceptions.internal.EventLogException;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import dev.fitko.fitconnect.api.exceptions.internal.SubmitEventNotFoundException;
import dev.fitko.fitconnect.api.exceptions.internal.ValidationException;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.client.subscriber.ReplySender;
import dev.fitko.fitconnect.client.subscriber.SubmissionReceiver;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.utils.EventLogUtil;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.Mockito;

import java.io.File;
import java.io.IOException;
import java.net.URI;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.ArgumentMatchers.anyList;
import static org.mockito.Mockito.anyLong;
import static org.mockito.Mockito.anyString;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

class SubscriberClientTest {

    private static final ObjectMapper MAPPER = new ObjectMapper().configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

    private FitConnectService subscriberMock;
    private RSAKey privateKey;
    private SubscriberClient underTest;
    private ReplySender replySenderMock;

    @BeforeEach
    public void setup() throws IOException, ParseException {
        final EnvironmentName environmentName = new EnvironmentName("TEST");
        final ApplicationConfig config = ApplicationConfig.builder()
                .environments(Map.of(environmentName, Environments.TEST.getEnvironment()))
                .activeEnvironment(environmentName)
                .build();

        privateKey = RSAKey.parse(getResourceAsString("private_decryption_test_key.json"));
        subscriberMock = Mockito.mock(FitConnectService.class);
        replySenderMock = Mockito.mock(ReplySender.class);
        underTest = new SubscriberClient(subscriberMock, new SubmissionReceiver(subscriberMock, privateKey, config), replySenderMock);
    }

    @Test
    void testNoSubmissionsFound() {

        final SubmissionsForPickup submissionsForPickup = new SubmissionsForPickup();
        submissionsForPickup.setTotalCount(0);

        final var destinationId = UUID.randomUUID();
        when(subscriberMock.getAvailableSubmissions(destinationId, 0, 25)).thenReturn(submissionsForPickup);
        assertThat(underTest.getAvailableSubmissionsForDestination(destinationId).isEmpty(), is(true));
    }

    @Test
    void testRestCallFailed() {
        // Given
        when(subscriberMock.getSubmission(any())).thenThrow(new RestApiException("Submission not found"));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(UUID.randomUUID()));

        // Then
        assertThat(exception.getMessage(), containsString("Submission not found"));
        assertThat(exception.getCause().getMessage(), containsString("Submission not found"));
    }

    @Test
    void testDecryptionFailed() {
        // Given
        final var submissionId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData("encryptedData");
        submission.setEncryptedMetadata("encryptedMetadata");

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTags");
        authenticationTags.setData("dataAuthTags");

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.decryptString(any(), any())).thenThrow(new DecryptionException("Decryption failed"));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submissionId));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("Decryption failed"));
    }

    @Test
    void testAcceptSubmission() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setCaseId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setEncryptedMetadata("abc");

        final var hash = new Hash();
        hash.setContent("hashedTestContent");

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);
        submissionSchema.setSchemaUri(URI.create("https://test-schema.fitko.de"));

        final var data = new Data();
        data.setHash(hash);
        data.setSubmissionSchema(submissionSchema);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("abc");
        authenticationTags.setData("def");

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptString(any(), any())).thenReturn(MAPPER.writeValueAsBytes(metadata));
        when(subscriberMock.validateSubmissionMetadata(metadata, submission, authenticationTags)).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.validateAttachments(any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateJsonFormat(any(), any())).thenReturn(ValidationResult.ok());

        // When
        underTest.requestSubmission(submissionId).acceptSubmission();

        // Then
        verify(subscriberMock, times(1)).acceptSubmission(EventPayload.forAcceptEvent(submission));
    }

    @Test
    void testUserRejectSubmission() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setCaseId(UUID.randomUUID());
        submission.setDestinationId(UUID.randomUUID());
        submission.setEncryptedMetadata("abc");

        final var hash = new Hash();
        hash.setContent("hashedTestContent");

        final var submissionSchema = new SubmissionSchema();
        submissionSchema.setMimeType(MimeType.APPLICATION_JSON);

        final var data = new Data();
        data.setHash(hash);
        data.setSubmissionSchema(submissionSchema);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);

        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("abc");
        authenticationTags.setData("def");

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptString(any(), any())).thenReturn(MAPPER.writeValueAsBytes(metadata));
        when(subscriberMock.validateSubmissionMetadata(metadata, submission, authenticationTags)).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.validateAttachments(any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateJsonFormat(any(), any())).thenReturn(ValidationResult.ok());

        // When
        underTest.requestSubmission(submissionId).rejectSubmission(List.of(new DataEncryptionIssue()));

        // Then
        final var expectedEventPayload = EventPayload.forRejectEvent(submission, List.of(new DataEncryptionIssue()));
        verify(subscriberMock, times(1)).rejectSubmission(expectedEventPayload);
    }

    @Test
    void testReadingMetadataFailed() {
        // Given
        final var submissionId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData("encryptedData");
        submission.setEncryptedMetadata("encryptedMetadata");

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("metadataAuthTags");
        authenticationTags.setData("dataAuthTags");

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.decryptString(any(), any())).thenReturn("encryptedMetadata".getBytes());

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submissionId));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("Unrecognized token 'encryptedMetadata'"));
    }

    @Test
    void testSubmissionsFound() {
        // Given
        final var destinationId = UUID.randomUUID();

        final SubmissionForPickup submission = new SubmissionForPickup();
        submission.setSubmissionId(UUID.randomUUID());
        submission.setDestinationId(destinationId);

        final SubmissionsForPickup submissionsForPickup = new SubmissionsForPickup();
        submissionsForPickup.setSubmissions(Set.of(submission));

        when(subscriberMock.getAvailableSubmissions(destinationId, 0, 1)).thenReturn(submissionsForPickup);

        // When
        final Set<SubmissionForPickup> availableSubmissions = underTest.getAvailableSubmissionsForDestination(destinationId, 0, 1);

        // Then
        assertThat(availableSubmissions.size(), is(1));
        assertThat(availableSubmissions.stream().findFirst().get(), is(submission));
    }

    @Test
    void testGetSubmission() throws Exception {

        // Given
        final var dataPayload = "{ some : 'data foo' }";
        final var submissionId = UUID.randomUUID();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Hash hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        hash.setSignatureType(SignatureType.SHA_512);

        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://dummy.schema.url"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(List.of());
        contentStructure.setData(data);

        final Metadata metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final byte[] metadataBytes = MAPPER.writeValueAsBytes(metadata);

        final RSAKey encryptionKey = privateKey;
        final RSAKey decryptionKey = encryptionKey.toPublicJWK();

        final String encryptedData = cryptoService.encryptObject(decryptionKey, dataPayload, "application/json");
        final String encryptedMetadata = cryptoService.encryptBytes(decryptionKey, metadataBytes, "application/json");

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);
        submission.setCallback(new Callback(URI.create("https://example.callback.net"), null));

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setData(encryptedData.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);
        authenticationTags.setMetadata(encryptedMetadata.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.validateSubmissionMetadata(metadata, submission, authenticationTags)).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateAttachments(any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateJsonFormat(any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.decryptString(encryptionKey, submission.getEncryptedData())).thenReturn(dataPayload.getBytes());
        when(subscriberMock.decryptString(encryptionKey, submission.getEncryptedMetadata())).thenReturn(metadataBytes);
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertNotNull(receivedSubmission);
        assertThat(receivedSubmission.getDataAsString(), is(dataPayload));
        assertThat(receivedSubmission.getDataMimeType(), is(MimeType.APPLICATION_JSON.value()));
        assertThat(receivedSubmission.getAttachments(), is(Collections.EMPTY_LIST));
        assertThat(receivedSubmission.getCallback().getUri(), is(URI.create("https://example.callback.net")));
    }

    @Test
    void testSubmissionNotFound() {
        // Given
        when(subscriberMock.getSubmission(any())).thenThrow(new RestApiException("Submission not found"));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(UUID.randomUUID()));

        // Then
        assertThat(exception.getMessage(), containsString("Submission not found"));
    }

    @Test
    void testMetadataValidationFailed() throws JsonProcessingException {
        // Given
        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setCaseId(caseId);
        submission.setEncryptedMetadata("abc");

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("abc");

        final var invalidMetadata = MAPPER.writeValueAsBytes(new Metadata());

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptString(any(), any())).thenReturn(invalidMetadata);
        when(subscriberMock.validateSubmissionMetadata(new Metadata(), submission, authenticationTags)).thenReturn(ValidationResult.error(new ValidationException("Metadata does not match schema")));
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);

        final IncorrectMetadataAuthenticationTag problem = new IncorrectMetadataAuthenticationTag();
        when(subscriberMock.validateSubmissionMetadata(new Metadata(), submission, authenticationTags)).thenReturn(ValidationResult.problem(problem));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submissionId));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("The authentication tag for the metadata is invalid"));
        verify(subscriberMock, times(1)).rejectSubmission(EventPayload.forRejectEvent(submission, List.of(problem)));
    }

    @Test
    void testUnknownPropertyOnMetadata() throws IOException {
        // Given
        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setCaseId(caseId);
        submission.setEncryptedMetadata("abc");

        final Map<String, Map<?, ?>> metadata = Map.of(
                "$schemaFoo", Collections.emptyMap(),
                "contentStructure", Map.of(
                        "data", Map.of(
                                "hash", Map.of(
                                        "type", SignatureType.SHA_512,
                                        "content", "foo"
                                ),
                                "submissionSchema", Map.of(
                                        "schemaUri", URI.create("https://dummy.schema.url"),
                                        "mimeType", MimeType.APPLICATION_JSON
                                )
                        )
                )
        );

        final var invalidMetadata = MAPPER.writeValueAsBytes(metadata);

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata("abc");
        authenticationTags.setData("def");

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptString(any(), any())).thenReturn(invalidMetadata);
        when(subscriberMock.validateSubmissionMetadata(MAPPER.readValue(invalidMetadata, Metadata.class), submission, authenticationTags)).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.validateAttachments(any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateJsonFormat(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertThat(receivedSubmission.getMetadata().getContentStructure().getData().getHash().getContent(), is("foo"));
    }

    @Test
    void testUnknownAdditionalReferenceInfoProperty() throws IOException {

        final var dataPayload = "{ some : 'data foo' }";
        final var submissionId = UUID.randomUUID();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Map<String, Map<String, Object>> metadata = Map.of(
                "contentStructure", Map.of(
                        "data", Map.of(
                                "hash", Map.of(
                                        "type", SignatureType.SHA_512,
                                        "content", cryptoService.hashBytes(dataPayload.getBytes())
                                ),
                                "submissionSchema", Map.of(
                                        "schemaUri", URI.create("https://dummy.schema.url"),
                                        "mimeType", MimeType.APPLICATION_JSON
                                )
                        ),
                        "attachments", Collections.EMPTY_LIST),
                "additionalReferenceInfo", Map.of(
                        "x-sender", "SOME_ADDITIONAL_UNKNOWN_PROPERTY",
                        "senderReference", "sender_123")
        );

        final byte[] metadataBytes = MAPPER.writeValueAsBytes(metadata);

        final RSAKey encryptionKey = privateKey;
        final RSAKey decryptionKey = encryptionKey.toPublicJWK();

        final String encryptedData = cryptoService.encryptObject(decryptionKey, dataPayload, "application/json");
        final String encryptedMetadata = cryptoService.encryptBytes(decryptionKey, metadataBytes, "application/json");

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setData(encryptedData.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);
        authenticationTags.setMetadata(encryptedMetadata.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.validateSubmissionMetadata(MAPPER.readValue(metadataBytes, Metadata.class), submission, authenticationTags)).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateJsonFormat(any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.decryptString(encryptionKey, submission.getEncryptedData())).thenReturn(dataPayload.getBytes());
        when(subscriberMock.decryptString(encryptionKey, submission.getEncryptedMetadata())).thenReturn(metadataBytes);
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.validateAttachments(any(), any())).thenReturn(ValidationResult.ok());

        // When
        final var receivedSubmission = underTest.requestSubmission(submissionId);

        // Then
        assertNotNull(receivedSubmission.getMetadata());
        assertThat(receivedSubmission.getMetadata().getAdditionalReferenceInfo().getSenderReference(), is("sender_123"));
    }

    @Test
    void testDataValidationFailed() throws JsonProcessingException {

        // Given
        final RSAKey decryptionKey = privateKey;
        final RSAKey encryptionKey = decryptionKey.toPublicJWK();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final var dataPayload = "some data that was transferred";

        final var hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        final var data = new Data();
        data.setHash(hash);
        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String encryptedData = cryptoService.encryptObject(encryptionKey, MAPPER.writeValueAsString(data), "application/json");
        final String encryptedMetadata = cryptoService.encryptObject(encryptionKey, MAPPER.writeValueAsString(metadata), "application/json");
        final var metadataBytes = MAPPER.writeValueAsBytes(metadata);
        final var dataBytes = MAPPER.writeValueAsBytes(data);

        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setCaseId(caseId);
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setEncryptedMetadata(encryptedMetadata);
        submission.setEncryptedData(encryptedData);

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setData(encryptedData.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);
        authenticationTags.setMetadata(encryptedMetadata.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptString(decryptionKey, encryptedMetadata)).thenReturn(metadataBytes);
        when(subscriberMock.decryptString(decryptionKey, encryptedData)).thenReturn(dataBytes);
        when(subscriberMock.validateSubmissionMetadata(metadata, submission, authenticationTags)).thenReturn(ValidationResult.ok());

        final DataHashMismatch problem = new DataHashMismatch();
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.problem(problem));
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submissionId));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("Submission data checksum wrong"));
        verify(subscriberMock, times(1)).rejectSubmission(EventPayload.forRejectEvent(submission, List.of(problem)));

    }

    @Test
    void testInvalidSubmissionDataSchema() throws JsonProcessingException {

        // Given
        final RSAKey decryptionKey = privateKey;
        final RSAKey encryptionKey = decryptionKey.toPublicJWK();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final var dataPayload = "some data that was transferred";

        final var hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        final SubmissionSchema submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("urn:something"));
        final var data = new Data();
        data.setHash(hash);
        data.setSubmissionSchema(submissionSchema);
        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String encryptedData = cryptoService.encryptObject(encryptionKey, MAPPER.writeValueAsString(data), "application/json");
        final String encryptedMetadata = cryptoService.encryptObject(encryptionKey, MAPPER.writeValueAsString(metadata), "application/json");
        final var metadataBytes = MAPPER.writeValueAsBytes(metadata);
        final var dataBytes = MAPPER.writeValueAsBytes(data);

        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setCaseId(caseId);
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setEncryptedMetadata(encryptedMetadata);
        submission.setEncryptedData(encryptedData);

        final AuthenticationTags authenticationTags = AuthenticationTags.from(encryptedData, encryptedMetadata, Collections.emptyMap());

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptString(decryptionKey, encryptedMetadata)).thenReturn(metadataBytes);
        when(subscriberMock.decryptString(decryptionKey, encryptedData)).thenReturn(dataBytes);
        when(subscriberMock.validateSubmissionMetadata(metadata, submission, authenticationTags)).thenReturn(ValidationResult.ok());
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.validateJsonFormat(any(), any())).thenReturn(ValidationResult.error(new RuntimeException("Something went wrong!")));

        final DataSchemaViolation problem = new DataSchemaViolation();
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.problem(problem));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submissionId));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("Submission data does not comply to schema"));
        verify(subscriberMock, times(1)).rejectSubmission(EventPayload.forRejectEvent(submission, List.of(problem)));
    }

    @Test
    void testAttachmentValidationFailed() throws JsonProcessingException {
        // Given
        final RSAKey decryptionKey = privateKey;
        final RSAKey encryptionKey = decryptionKey.toPublicJWK();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final var dataHash = new Hash();
        dataHash.setContent("");
        final SubmissionSchema submissionSchema = new SubmissionSchema();
        submissionSchema.setSchemaUri(URI.create("urn:something"));
        final var data = new Data();
        data.setHash(dataHash);
        data.setSubmissionSchema(submissionSchema);

        final var attachmentHash = new Hash();
        attachmentHash.setContent("");

        final var attachment = new ApiAttachment();
        attachment.setAttachmentId(UUID.randomUUID());
        attachment.setFilename("src/test/resources/attachment.txt");
        attachment.setHash(attachmentHash);

        final var contentStructure = new ContentStructure();
        contentStructure.setData(data);
        contentStructure.setAttachments(List.of(attachment));
        final var metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final String encryptedData = cryptoService.encryptBytes(encryptionKey, MAPPER.writeValueAsString(data).getBytes(), "application/json");
        final String encryptedMetadata = cryptoService.encryptBytes(encryptionKey, MAPPER.writeValueAsString(metadata).getBytes(), "application/json");
        final var metadataBytes = MAPPER.writeValueAsBytes(metadata);
        final var dataBytes = MAPPER.writeValueAsBytes(data);

        final var submissionId = UUID.randomUUID();
        final var destinationId = UUID.randomUUID();
        final var caseId = UUID.randomUUID();

        final var submission = new Submission();
        submission.setCaseId(caseId);
        submission.setSubmissionId(submissionId);
        submission.setDestinationId(destinationId);
        submission.setEncryptedMetadata(encryptedMetadata);
        submission.setEncryptedData(encryptedData);

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setData(encryptedData.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);
        authenticationTags.setMetadata(encryptedMetadata.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.decryptString(decryptionKey, encryptedMetadata)).thenReturn(metadataBytes);
        when(subscriberMock.decryptString(decryptionKey, encryptedData)).thenReturn(dataBytes);
        when(subscriberMock.validateSubmissionMetadata(metadata, submission, authenticationTags)).thenReturn(ValidationResult.ok());
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateJsonFormat(any(), any())).thenReturn(ValidationResult.ok());

        final IncorrectAttachmentAuthenticationTag problem = new IncorrectAttachmentAuthenticationTag(attachment.getAttachmentId());
        when(subscriberMock.validateAttachments(anyList(), any())).thenReturn(ValidationResult.problem(problem));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submissionId));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("The authentication tag for the attachment " + attachment.getAttachmentId() + " is invalid"));
        verify(subscriberMock, times(1)).rejectSubmission(EventPayload.forRejectEvent(submission, List.of(problem)));
    }

    @Test
    void testMissingAttachment() throws Exception {

        // Given
        final var dataPayload = "{ some : 'data foo' }";
        final var submissionId = UUID.randomUUID();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Hash hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        hash.setSignatureType(SignatureType.SHA_512);

        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://dummy.schema.url"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ApiAttachment attachment = new ApiAttachment();
        attachment.setAttachmentId(UUID.randomUUID());

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(List.of(attachment));
        contentStructure.setData(data);

        final Metadata metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final byte[] metadataBytes = MAPPER.writeValueAsBytes(metadata);

        final RSAKey encryptionKey = privateKey;
        final RSAKey decryptionKey = encryptionKey.toPublicJWK();

        final String encryptedData = cryptoService.encryptBytes(decryptionKey, dataPayload.getBytes(StandardCharsets.UTF_8), "application/json");
        final String encryptedMetadata = cryptoService.encryptBytes(decryptionKey, metadataBytes, "application/json");

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata(encryptedMetadata.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.validateSubmissionMetadata(metadata, submission, authenticationTags)).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.decryptString(encryptionKey, submission.getEncryptedData())).thenReturn(dataPayload.getBytes());
        when(subscriberMock.decryptString(encryptionKey, submission.getEncryptedMetadata())).thenReturn(metadataBytes);
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.getSubmissionAttachment(any(), any())).thenThrow(new RestApiException("Attachment download failed", 404));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submissionId));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("Attachment download failed"));
        verify(subscriberMock, times(1)).rejectSubmission(EventPayload.forRejectEvent(submission, List.of(new MissingAttachment(attachment.getAttachmentId()))));
    }

    @Test
    void testAttachmentEncryptionIssue() throws Exception {

        // Given
        final var dataPayload = "{ some : 'data foo' }";
        final var submissionId = UUID.randomUUID();
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Hash hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        hash.setSignatureType(SignatureType.SHA_512);

        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://dummy.schema.url"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ApiAttachment attachment = new ApiAttachment();
        attachment.setAttachmentId(UUID.randomUUID());

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(List.of(attachment));
        contentStructure.setData(data);

        final Metadata metadata = new Metadata();
        metadata.setContentStructure(contentStructure);

        final byte[] metadataBytes = MAPPER.writeValueAsBytes(metadata);

        final RSAKey encryptionKey = privateKey;
        final RSAKey decryptionKey = encryptionKey.toPublicJWK();

        final String encryptedData = cryptoService.encryptBytes(decryptionKey, dataPayload.getBytes(StandardCharsets.UTF_8), "application/json");
        final String encryptedMetadata = cryptoService.encryptBytes(decryptionKey, metadataBytes, "application/json");

        final var submission = new Submission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);

        final AuthenticationTags authenticationTags = new AuthenticationTags();
        authenticationTags.setMetadata(encryptedMetadata.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);

        when(subscriberMock.getSubmission(any())).thenReturn(submission);
        when(subscriberMock.validateSubmissionMetadata(metadata, submission, authenticationTags)).thenReturn(ValidationResult.ok());
        when(subscriberMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.ok());
        when(subscriberMock.decryptString(encryptionKey, submission.getEncryptedData())).thenReturn(dataPayload.getBytes());
        when(subscriberMock.decryptString(encryptionKey, submission.getEncryptedMetadata())).thenReturn(metadataBytes);
        when(subscriberMock.decryptString(encryptionKey, "encrypt$dAtt@chm$nt")).thenThrow(new DecryptionException("Decrypting attachment failed"));
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenReturn(authenticationTags);
        when(subscriberMock.getSubmissionAttachment(any(), any())).thenReturn("encrypt$dAtt@chm$nt");

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submissionId));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("Decrypting attachment failed"));
        verify(subscriberMock, times(1)).rejectSubmission(EventPayload.forRejectEvent(submission, List.of(new AttachmentEncryptionIssue(attachment.getAttachmentId()))));
    }

    @Test
    void testNotExactlyOneSubmitEvent() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());

        final var expectedProblem = new NotExactlyOneSubmitEvent();

        when(subscriberMock.getSubmission(submission.getSubmissionId())).thenReturn(submission);
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenThrow(new SubmitEventNotFoundException("no submit event in log"));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submission.getSubmissionId()));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("The Event-Log is inconsistent because it does not contain exactly one 'submit' event."));
        verify(subscriberMock, times(1)).rejectSubmission(EventPayload.forRejectEvent(submission, List.of(expectedProblem)));
    }

    @Test
    void testMissingAuthenticationTags() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());

        final var expectedProblem = new MissingAuthenticationTags();

        when(subscriberMock.getSubmission(submission.getSubmissionId())).thenReturn(submission);
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenThrow(new AuthenticationTagsEmptyException("empty auth tags"));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submission.getSubmissionId()));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("The 'submit-submission' event does not contain authentication tags"));
        verify(subscriberMock, times(1)).rejectSubmission(EventPayload.forRejectEvent(submission, List.of(expectedProblem)));
    }

    @Test
    void testInvalidEventLog() {

        // Given
        final var submission = new Submission();
        submission.setSubmissionId(UUID.randomUUID());

        final InvalidEventLog expectedProblem = new InvalidEventLog();

        when(subscriberMock.getSubmission(submission.getSubmissionId())).thenReturn(submission);
        when(subscriberMock.getSubmissionAuthenticationTags(submission)).thenThrow(new EventLogException("The Event-Log is inconsistent"));

        // When
        final FitConnectSubscriberException exception = assertThrows(FitConnectSubscriberException.class, () -> underTest.requestSubmission(submission.getSubmissionId()));

        // Then
        assertThat(exception.getCause().getMessage(), containsString("The Event-Log is inconsistent"));
        verify(subscriberMock, times(1)).rejectSubmission(EventPayload.forRejectEvent(submission, List.of(expectedProblem)));
    }

    @Test
    void validateCallback() {

        when(subscriberMock.validateCallback(anyString(), anyLong(), anyString(), anyString())).thenReturn(ValidationResult.ok());

        final ValidationResult validationResult = underTest.validateCallback("hmac", 0L, "body", "secret");

        verify(subscriberMock, times(1)).validateCallback(anyString(), anyLong(), anyString(), anyString());
        assertTrue(validationResult.isValid());
        assertFalse(validationResult.hasError());
    }

    private String getResourceAsString(final String name) throws IOException {
        final ClassLoader classLoader = getClass().getClassLoader();
        final File file = new File(Objects.requireNonNull(classLoader.getResource(name)).getFile());
        return Files.readString(file.toPath());
    }
}